/*
 * \brief  Main program of the Hello server
 * \author Björn Döbel
 * \date   2008-03-20
 */

/*
 * Copyright (C) 2008-2009
 * Genode Labs, Feske & Helmuth Systementwicklung GbR
 *
 * This file is part of the Genode OS framework, which is distributed
 * under the terms of the GNU General Public License version 2.
 */

#include <base/printf.h>
#include <base/env.h>
#include <base/sleep.h>
#include <base/thread.h>
#include <cpu_session/cpu_session.h>
#include <cap_session/connection.h>
#include <sampling_session/connection.h>
#include <root/component.h>

#include <sampling_session/component.h>
#include "my_threads.h"
#include "profile_port.h"

namespace Okl4 {extern "C" {
#include <l4/profpolicy.h>
#include <l4/arch/syscalls.h>
#include <l4/thread.h>
}}

namespace Sampling {

	class Root_component : public Genode::Root_component<Sampling::Session_component>
	{
		protected:

			Sampling::Session_component *_create_session(const char *args)
			{
				PDBG("creating sampling session.\n");
				return new (md_alloc()) Sampling::Session_component();
			}

		public:

			Root_component(Genode::Server_entrypoint *ep,
			               Genode::Allocator         *allocator)
			: Genode::Root_component<Sampling::Session_component>(ep, allocator)
			{
				PDBG("Sampling: Creating root component.");
			}
	};
}

using namespace Genode;
using namespace Okl4;

Sampling::Session_capability *h_cap;
Sampling::Session_client *h;

int main(void)
{
	/* 
	 * Get a session for the parent's capability service, so that we are able
	 * to create capabilities.
	 */
	Cap_connection cap;

	/*
	 * A sliced heap is used for allocating session objects - thereby we can
	 * release objects separately.
	 */
	static Sliced_heap sliced_heap(env()->ram_session(), env()->rm_session());

	/*
	 * Create objects for use by the framework.
	 *
	 * Msgbufs are used for sending and receiving messages using a
	 * 'Server_activation'. A 'Server_entrypoint' is created to announce
	 * our service's root capability to our parent and manage incoming
	 * session creation requests.
	 */
	static Msgbuf<256> snd_msg, rcv_msg;
	static Server_activation<4096> act(&snd_msg, &rcv_msg);
	static Server_entrypoint entry(&cap, &act);

	static Sampling::Root_component sampling_root(&entry, &sliced_heap);
	env()->parent()->announce(Sampling::Session::service_name(),
	                          Root_capability(entry.manage(&sampling_root)));

	Sampling::Session_capability h_cap(env()->parent()->session(
								Sampling::Session::service_name(), "foo, ram_quota=4K"));
	Sampling::Session_client h(h_cap);

	Dataspace_capability ds_cap = h.get_cap();
	volatile void *ptr = Genode::env()->rm_session()->attach(ds_cap);

	*((int*)ptr) = 1000;
	
	printf("SERVER MEM VALUE CONTENTS IS %d\n", *((int*)ptr));

	init_matrix();
	type_table_init();
	
	prof_interrupt_handler handler;
	handler.start();
	registration_thread r_thread;
	r_thread.start();
	profiling_links p_links;
	p_links.start();
	//profiling_init();
	//netsamp_thread nthread;
	//nthread.start();
	//profiling_init();
	/* We are done with this and only act upon client requests now. */

	sleep_forever();

	return 0;
}
