#include "my_hash.h"
#include <base/printf.h>
#include <util/string.h>
#include <base/env.h>

/*
	hsequence: stores the type sequence as a string (null-terminated)
	vsequence: stores the value sequence as a string (null-terminated)
*/

#define NULL ((void *)0)
using namespace Genode;

struct hash_table *init_hash_table(char *hsequence, char *vsequence)
{
	struct hash_table *table;
	
	table = new(Genode::env()->heap()) struct hash_table[1];
	
	if(!table)
	{
		printf("No hash table created\n");
		return (struct hash_table*)NULL;
	}
	
	struct hash_entry *hentry_ptr = (struct hash_entry*)(table);
	
	int loop;
	for(loop = 0;loop < 48;loop++)
	{//we know a hash element hasn't been written if the buffers storing the
	 //hashes and values are null.
		table->entries[loop] = *hentry_ptr;
		table->entries[loop].hash_sequence = (char*)NULL;
		table->entries[loop].value_sequence = (char*)NULL;
		table->entries[loop].next = (struct hash_entry*)NULL;
		hentry_ptr++;
	}
	
	//I might have to include the hash sequence type in the hash entry....
	table->hash_types = new(Genode::env()->heap()) char[Genode::strlen(hsequence) + 1];
	table->value_types = new(Genode::env()->heap()) char[Genode::strlen(vsequence) + 1];
	Genode::memcpy(table->hash_types, hsequence, Genode::strlen(hsequence));
	Genode::memcpy(table->value_types, vsequence, Genode::strlen(vsequence));
	*(table->hash_types + Genode::strlen(hsequence) + 1) = '\0';
	*(table->value_types + Genode::strlen(vsequence) + 1) = '\0';
    return table;
}

char* lookup_hash(struct hash_table *table, char *sequence)
{
    uintptr_t hash;
    struct hash_entry *entry;

	hash = seq_to_hash(sequence, hash_memory(table, sequence)) & 47;

	//if(&table->entries[hash] == NULL)
		//printf("This shouldn't appear at all\n");

    for (entry = &table->entries[hash]; entry != NULL; entry = entry->next) {
    	if(entry->hash_sequence != NULL && entry->value_sequence != NULL)
    	{
	        if (same_key_sequence(table, entry->hash_sequence, sequence)) {
	            return entry->value_sequence;
	        }
	    }
    }
    return (char*)NULL;
}

int same_key_sequence(struct hash_table *table, char *sequence1, char *sequence2)
{
	char *hash_seq = table->hash_types;
	char *seq1 = sequence1;
	char *seq2 = sequence2;
	char type_value = 0;
	
	if(sequence1 == NULL || sequence2 == NULL)
		return 0;
	
	unsigned int i;
	
	for(i = 0;i < Genode::strlen(hash_seq);i++)
	{
		type_value = table->hash_types[i];
		
		switch(type_value)
		{
			case INT:
				{
					int first = *((int*)seq1);
					int second = *((int*)seq2);
					
					if(first != second)
						return 0;
					
					seq1 += sizeof(int);
					seq2 += sizeof(int);
				}
				break;
			case DOUBLE:
				{
					double first = *((double*)seq1);
					double second = *((double*)seq2);
					
					if(first != second)
						return 0;
						
					seq1 += sizeof(double);
					seq2 += sizeof(double);
				}
				break;
			case STRING:
				{
					if(Genode::strcmp(seq1, seq2) != 0)
						return 0;
					
					seq1 += Genode::strlen(seq1) + 1;
					seq2 += Genode::strlen(seq2) + 1;
				}
				break;
			//case L4_WORD:
			//case TID:
			/*case UINT:
				{
					unsigned long first = *((unsigned long*)seq1);
					unsigned long second = *((unsigned long*)seq2);
					
					if(first != second)
						return 0;
					
					seq1 += sizeof(unsigned long);
					seq2 += sizeof(unsigned long);
				}
				break;*/
		}
	}

	return 1;
}

int hash_value_size(struct hash_table *table, char *value_sequence)
{
	char *hash_seq = table->hash_types;
	char *seq1 = value_sequence;
	char type_value = 0;
	int value_size = 0;
	
	unsigned int i;
	for(i = 0;i < Genode::strlen(hash_seq);i++)
	{
		type_value = table->value_types[i];
		
		switch(type_value)
		{
			//case UINT:
			case INT:
				value_size += sizeof(int);
				seq1 += sizeof(int);
				break;
			case DOUBLE:
				value_size += sizeof(double);
				seq1 += sizeof(double);
				break;
			case STRING:
				value_size += Genode::strlen(seq1) + 1;
				seq1 += Genode::strlen(seq1) + 1;
				break;
		}
	}
	
	return value_size;
}

/* The string hash being used is Bob Jenkin's one-at-a-time algorithm */
uintptr_t seq_to_hash(char *sequence, unsigned long int key_len)
{
    uintptr_t hash = 0;
    Genode::size_t i;
 
    for (i = 0; i < key_len; i++) {
        hash += (unsigned)sequence[i];
        hash += (hash << 10);
        hash ^= (hash >> 6);
    }
    hash += (hash << 3);
    hash ^= (hash >> 11);
    hash += (hash << 15);
    return hash;
}

int hash_memory(struct hash_table *table, char *sequence)
{
	char *reference = sequence;
	char type_value = 0;
	int size = 0;

	unsigned int i;	
	for(i = 0;i < Genode::strlen(table->hash_types);i++)
	{
		type_value = table->hash_types[i];
		
		switch(type_value)
		{//NEED TO ADD BOOLEAN CASE
			case INT:
				size += sizeof(int);
				reference += sizeof(int);
				break;
			case DOUBLE:
				size += sizeof(double);
				reference += sizeof(double);
				break;
			case STRING:
				size += Genode::strlen(reference) + 1;
				reference += Genode::strlen(reference) + 1;
				break;
		}
	}
	
	return size;
}

int insert_hash(struct hash_table *table, char *sequence, char *value)
{
    uintptr_t hash;
    struct hash_entry *entry;

	hash = seq_to_hash(sequence, hash_memory(table, sequence)) & 47;
	struct hash_entry *ptr = &table->entries[hash];
	int value_size = hash_value_size(table, sequence);

	if(ptr->hash_sequence == NULL) //yes: value_sequence has to be null as well.
	{//this hasn't been assigned yet, so use the current memory.   
		ptr->hash_sequence = new(Genode::env()->heap()) char[hash_memory(table, sequence)];
		ptr->value_sequence = new(Genode::env()->heap()) char[value_size];

		if(!ptr->hash_sequence || sequence == NULL)
			return -1;
			
		Genode::memcpy(ptr->value_sequence, value, value_size);
		Genode::memcpy(ptr->hash_sequence, sequence, hash_memory(table, sequence));	
		ptr->next = (struct hash_entry*)NULL;
	}
    else
    {
    	//iterate through current elements until we get the entry that has
    	//a NULL pointer. Allocate memory for a new hash entry, and attach
    	//that to the hash table. Also assign appropriate values as necessary

    	while(ptr->next != NULL)
    	{
    		ptr = ptr->next;
    	}

	    entry = new(Genode::env()->heap()) struct hash_entry[1];
	    if (!entry)
	    	return -1;

	    entry->value_sequence = new(Genode::env()->heap()) char[value_size];
	    entry->hash_sequence = new(Genode::env()->heap()) char[hash_memory(table, sequence)];
	    Genode::memcpy(entry->value_sequence, value, value_size);
		Genode::memcpy(entry->hash_sequence, sequence, hash_memory(table, sequence));	

		entry->next = (struct hash_entry*)NULL;
		ptr->next = entry;

    }
    
    return 0;
}

char* hash_value_location(struct hash_table *table, char *sequence)
{
    uintptr_t hash;
    struct hash_entry *entry;

	hash = seq_to_hash(sequence, hash_memory(table, sequence)) & 47;

    for (entry = &table->entries[hash]; entry != NULL; entry = entry->next) {
        if (same_key_sequence(table, entry->value_sequence, sequence)) {
            return entry->value_sequence;
        }
    }

    return (char*)NULL;
}

void hash_element_print(struct hash_table *table)
{
	int counter;
    struct hash_entry *entry;
    
	printf("Key, Value:\n");
    for (counter = 0; counter < 48; counter++) 
    {
        entry = &table->entries[counter];
        //printf("value of counter is %d\n", counter);
        while (entry) 
        {
        	if(entry->hash_sequence != NULL)
        	{
    			print_hash_key(table, entry);
        		printf("\t\t\t%d\n", *((int*)entry->value_sequence));
        	}
	        
            entry = entry->next;
        }
    }
}

void print_hash_key(struct hash_table *table, struct hash_entry *entry)
{
	char type_value = 0;
	char *reference = entry->hash_sequence;
	printf("<");
	
	unsigned int i;
	for(i = 0;i < Genode::strlen(table->hash_types);i++)
	{
		type_value = table->hash_types[i];
		
		switch(type_value)
		{
			case INT:
				printf("%ld", *((long*)reference));
				reference += sizeof(int);
				break;
			case DOUBLE:
				printf("%f", *((double*)reference));
				reference += sizeof(double);
				break;
			case STRING:
				printf("%s", reference);
				reference += Genode::strlen(reference) + 1;
				break;
			case BOOLEAN:
				printf("%c", *reference);
				reference += 1;
				break;
		}
		
		if(i != Genode::strlen(table->hash_types) - 1)
			printf(",");
	}
	printf(">");
}
