#include "profile_port.h"

/**************************************************************************
	Standard C function implementations
**************************************************************************/

void assert(bool cond)
{
	if(!cond)
	{
		L4_KDB_Enter("Assertion failed\n");
	}
}
using namespace Genode;
void *operator new (Genode::size_t size)
{
	return env()->heap()->alloc(size);
}

char* realloc(char *old_buffer, int old_size, int new_size)
{
	char *new_buffer = new(env()->heap())char[new_size];
	Genode::memcpy((void*)new_buffer, (void*)old_buffer, old_size);
	destroy(Genode::env()->heap(), old_buffer);
	return new_buffer;
}
/**************************************************************************/

volatile int matrix_index;
volatile int thread_matrix[THREAD_QUANTITY][THREAD_QUANTITY];
volatile buffer_info *thread_buffers[THREAD_QUANTITY];

volatile L4_ThreadId_t thread_tunit[7];
L4_ThreadId_t list_of_reg_threads[THREAD_QUANTITY];

uintptr_t thread_ptrs[THREAD_QUANTITY];
L4_Word_t thread_list[THREAD_QUANTITY];
volatile L4_Word_t thread_list_index;

//Genode::network_buffer *network_buf;

void init_matrix()
{	
	int i, j;
	for(i = 0;i < THREAD_QUANTITY;i++)
	{
		for(j = 0;j < THREAD_QUANTITY;j++)
			thread_matrix[i][j] = 0;
	}

	for(i = 0;i < 7;i++)
		thread_tunit[i].raw = 0;
}


void assign_tid_tunit(L4_ThreadId_t coll_thread, int unit_assign)
{
	thread_tunit[unit_assign].raw = coll_thread.raw;
}

/* Get the index of the thread ID passed */
int thread_index(L4_ThreadId_t input)
{
	int i;
	for(i = 0;i < matrix_index;i++)
	{
		if(thread_list[i] == input.raw)
		{
			return i;
		}
	}

	return -1;
}

L4_Word_t tunit_index(int tunit)
{
	return thread_index(thread_tunit[tunit]);
}

L4_Word_t thread_ptr_index(uintptr_t thread_ptr)
{
	int i;
	for(i = 0;i < matrix_index;i++)
	{
		if(thread_ptrs[i] == thread_ptr)
			return i;
	}

	return -1;
}

L4_ThreadId_t thread_identifier(int index)
{
	if (index >= matrix_index)
		return L4_nilthread;

	return list_of_reg_threads[index];
}

int thread_index_fn(void (*fn)(void*))
{
	int i;
	for(i = 0;i < matrix_index;i++)
	{

		if(thread_ptrs[i] == (uintptr_t)fn)
		{
			return i;
		}
	}

	return -1;
}

buffer_info* binfo_raw_thread(L4_Word_t id)
{
	int i;
	for(i = 0;i < matrix_index;i++)
	{//Will need to inspect matrix and check whether the thread
	 //is allowed to collect information about the buffer.
		if(thread_list[i] == id)
			return (buffer_info*)thread_buffers[i];
	}
	
	return NULL;
}

buffer_info* thread_buffer_info(int index)
{
	if(index < matrix_index)
	{
		return (buffer_info*)thread_buffers[index];
	}

	return NULL;
}

int buffer_info_index(buffer_info *info)
{
	int i;
	for(i = 0;i < matrix_index;i++)
	{
		if((uintptr_t)thread_buffers[i] == (uintptr_t)info)
			return i;
	}
	
	return -1;
}

buffer_info* binfo_fn(uintptr_t fn)
{
	int i;
	for(i = 0;i < matrix_index;i++)
	{
		if(thread_ptrs[i] == fn)
			return thread_buffer_info(i);
	}

	return NULL;
}

L4_Word_t determine_size(buffer_info *info)
{//this should only apply to a collection spec
	//use this to allocate memory for the core window
	
	int sample_size = info->actual_chunk_size;

	/*
		If the time units for the sample collection and consolidation are the same,
		the amount of memory to allocate is always going to be:
		
		window length * sample size * number of windows
	*/
	if((info->interrupt_no == MSEC_TIMER && info->window_unit == UNIT_MSEC)
		|| (info->interrupt_no == SEC_TIMER && info->window_unit == UNIT_SEC)
		|| (info->interrupt_no == MIN_TIMER && info->window_unit == UNIT_MIN)
		|| (info->interrupt_no == HR_TIMER && info->window_unit == UNIT_HR))
	{
		return sample_size * info->time_window * info->no_windows;
	}
	else if(info->interrupt_no == MSEC_TIMER)
	{
		if(info->window_unit == UNIT_SEC)
			return sample_size * 513 * info->time_window * info->no_windows;
		else if(info->window_unit == UNIT_MIN)
			return sample_size * 30780 * info->time_window * info->no_windows;
		else if(info->window_unit == UNIT_HR)
			return sample_size * 1846800 * info->time_window * info->no_windows;
	}
	else if(info->interrupt_no == SEC_TIMER)
	{
		if(info->window_unit == UNIT_MSEC)
			return -1; 
		else if(info->window_unit == UNIT_MIN)
			return sample_size * 60 * info->time_window * info->no_windows;
		else if(info->window_unit == UNIT_HR)
			return sample_size * 3600 * info->time_window * info->no_windows;
	}
	else if(info->interrupt_no == MIN_TIMER)
	{
		if(info->window_unit == UNIT_MSEC || info->window_unit == UNIT_SEC)
			return -1;
		else if(info->window_unit == UNIT_HR)
			return sample_size * 60 * info->time_window * info->no_windows;
	}
	else if(info->interrupt_no == HR_TIMER)
	{
		if(info->window_unit == UNIT_MSEC || info->window_unit == UNIT_SEC || info->window_unit == UNIT_HR)
			return -1;
	}
	
	return -1;
}

L4_Word_t sample_equals(buffer_info *info, char *sample1, char *sample2)
{
	/* 
		Sample equality is defined as the following:
			- same interrupt value
			- same information collected
			- same TID and/or inst, depending on what is collected
	*/
	
	//check each element -> only check either TID, instruction addr or both
	
	//if(L4interrupt_type(sample1) == L4interrupt_type(sample2) &&
	//	L4info_collect(sample1) == L4info_collect(sample2) &&
	//	L4sample_size(sample1) == L4sample_size(sample2))
	if(info->info_collect & INFO_TID && info->info_collect & INFO_INST)
	{//if so, we are collecting both TID and instruction addr info
		int first_1 = *((int*)sample1);
		int first_2 = *((int*)sample2);
		unsigned int second_1 = *((int*)(sample1 + 4));
		unsigned int second_2 = *((int*)(sample2 + 4));

		if(first_1 == first_2 && second_1 == second_2)
		{
			return 1;
		}
	}
	else if(info->info_collect & INFO_TID || info->info_collect & INFO_INST)
	{
		int first_1 = *((int*)sample1);
		int first_2 = *((int*)sample2);

		if(first_1 == first_2)
			return 1;
	}

	return 0;
	
}

uintptr_t minimum_index_bounds(buffer_info *buffer)
{
	int non_null_pointers = 0;
	int null_char_pointers = 0;
	int scanned = 0;
	
	L4_Word_t min = 4294967295U;
	
	/*
		To obtain the minimum index, the value has to lie in between the write index and the end
		of the buffer.  In the case where all related indices are not within the bounds, this
		would indicate that we are in the starting copying state (ie. where samples can be written
		normally).
	*/
	
	int i;
	for(i = 0;i < matrix_index;i++)
	{
		if(buffer->moved[i])
		{
			scanned = 1;
			non_null_pointers++;
			
			if(buffer->t_offsets[i] < min)
				//&& buffer->buffer + buffer->t_offsets[i] > (uintptr_t)buffer->write_index)
			{
				min = buffer->t_offsets[i];
			}
		}
	}
	
	/* Returned if there are no thread links at all */
	if(null_char_pointers == matrix_index)
		return NO_THREAD_LINKS;
	
	/* Returned if there are thread links and the threads in question have
		not started reading data yet */
	if(non_null_pointers == buffer->connected_threads && scanned == 0)
		return NO_BUFFER_SCANS;
	
	return min;
}

void set_reset_state(buffer_info *info)
{
	int i = 0;
	for(i = 0;i < matrix_index;i++)
	{
		if(info->threads[i] != NULL)
		{
			info->resetted[i] = 1;
		}
	}
}

int memory_single_chunk(char *sequence)
{
	int i;
	int size = 0;
	int length = (int)Genode::strlen(sequence);
	for(i = 0;i < length;i++)
	{
		switch(sequence[i])
		{//NEED TO ADD BOOLEAN CASE
			case INT:
				size += sizeof(int);
				break;
			case DOUBLE:
				size += sizeof(double);
				break;
			case BOOLEAN:
				size += 1;
				break;
			case STRING:
				size += 50;
				break;
		}
	}
	return size;
}

//void finish_stream(L4_ThreadId_t tid)
void finish_stream(void (*fn)(void*))
{
	//buffer_info *info = binfo_raw_thread(tid.raw);
	buffer_info *info = binfo_fn((uintptr_t)fn);
	L4_Word_t min_index = minimum_index_bounds(info);
	int first_time = 0;
	
	if(info->reset == 0)
	{
		if(info->write_index + 4 >= info->buffer + info->size)
		{
			if(min_index != NO_THREAD_LINKS && min_index != NO_BUFFER_SCANS &&
				info->buffer + 4 < info->buffer + min_index)
			{
				first_time = 1;
			}

			set_reset_state(info);
			info->reset = 1;
		}
		else
		{
			char *end = (char*)"END";
			Genode::memcpy((void*)info->write_index, end, Genode::strlen(end) + 1);
			info->write_index += 4;
		}
	}
	
	if(info->reset == 1)
	{
		if(first_time == 1)
		{
			first_time = 0;
			*(info->write_index) = BUFFER_RESET;
			info->write_index = info->buffer;
		}
		else if(info->write_index + 1 < info->buffer + min_index)
		{
			//*(info->write_index) = END_OF_RECORDS;
			Genode::memcpy((void*)info->write_index, "END", Genode::strlen("END") + 1);
			info->write_index += 4;
		}
		else
		{
			info->dropped_samples++;
		}
	}
}

L4_ThreadId_t create_temp_id(L4_Word_t raw_id)
{
	L4_ThreadId_t new_id;
	new_id.raw = raw_id;
	
	return new_id;
}

#if 0
void suspend(void (*fn)(void*))
{	
	//get buffer_info for thread, if = 0, it must be a processing thread
	//L4_Word_t tid_search = thread_ptr_index((uintptr_t)fn);
	L4_Word_t tid_search = thread_ptr_index((uintptr_t)fn);
	buffer_info *tbuffer = thread_buffer_info(tid_search);

	if(tbuffer == NULL)
	{
		L4_ThreadId_t thr_suspend;
		thr_suspend.raw = thread_identifier(tid_search).raw;
		L4_Stop_Thread(thr_suspend);
	}
	else if(tbuffer->interrupt_no == 0)
	{
		//it is not a probe, so use the default scheduler
		L4_ThreadId_t thr_suspend;
		thr_suspend.raw = thread_identifier(tid_search).raw;
		L4_Stop_Thread(thr_suspend);
	}
	else
	{
		interrupt_spec tid_spec;
		tid_spec.tid = thread_identifier(tid_search).raw;
		L4_ProfPolicy(SAMPLING_SUSPEND, &tid_spec);
	}
}

void resume(void (*fn)(void*))
{
	L4_Word_t tid_search = thread_ptr_index((uintptr_t)fn);
	buffer_info *tbuffer = thread_buffer_info(tid_search);
	
	if(tbuffer->interrupt_no == 0)
	{
		//it is not a probe, so use the default scheduler
		L4_ThreadId_t thr_resume;
		thr_resume.raw = thread_identifier(tid_search).raw;
		L4_Start(thr_resume);
	}
	else
	{
		interrupt_spec tid_spec;
		tid_spec.tid = thread_identifier(tid_search).raw;
		L4_ProfPolicy(SAMPLING_RESUME, &tid_spec);
	}
}

void kill(void (*fn)(void*))
{
	L4_Word_t tid_search = thread_ptr_index((uintptr_t)fn);
	buffer_info *tbuffer = thread_buffer_info(tid_search);
	
	if(tbuffer->interrupt_no == 0)
	{
		//it is not a probe, so use the default scheduler
		L4_ThreadId_t thr_kill;
		thr_kill.raw = thread_identifier(tid_search).raw;
		//L4_ThreadControl(thr_kill, L4_nilspace, NULL, NULL, NULL, NULL, 0);
	}
	else
	{
		interrupt_spec tid_spec;
		tid_spec.tid = thread_identifier(tid_search).raw;
		L4_ProfPolicy(SAMPLING_KILL, &tid_spec);
	}
}
#endif

void suspend(ProfVector<L4_ThreadId_t> *vec)
{	
	//get buffer_info for thread, if = 0, it must be a processing thread
	//L4_Word_t tid_search = thread_ptr_index((uintptr_t)fn);
	//L4_Word_t tid_search = thread_ptr_index((uintptr_t)fn);
	
	buffer_info *tbuffer;
	
	int i;
	for(i = 0;i < vec->length;i++)
	{
		tbuffer = binfo_raw_thread(vec->buffer[i].raw);

		if(tbuffer == NULL)
		{
			L4_ThreadId_t thr_suspend;
			thr_suspend.raw = vec->buffer[i].raw;
			L4_Stop_Thread(thr_suspend);
		}
		else if(tbuffer->interrupt_no == 0)
		{
			//it is not a probe, so use the default scheduler
			L4_ThreadId_t thr_suspend;
			thr_suspend.raw = vec->buffer[i].raw;
			L4_Stop_Thread(thr_suspend);
		}
		else
		{
			interrupt_spec tid_spec;
			tid_spec.tid = vec->buffer[i].raw;
			
			switch(tbuffer->interrupt_no)
			{
				case MSEC_TIMER:
					thread_tunit[TID_MSEC].raw = 0;
					break;
				case SEC_TIMER:
					thread_tunit[TID_SEC].raw = 0;
					break;
				case MIN_TIMER:
					thread_tunit[TID_MIN].raw = 0;
					break;
				case HR_TIMER:
					thread_tunit[TID_HR].raw = 0;
					break;
				case PMC1:
					thread_tunit[TID_PMC1].raw = 0;
					break;
				case PMC2:
					thread_tunit[TID_PMC2].raw = 0;
					break;
			}
			
//			if(tbuffer->interrupt_no == HTTP_SERVER)
//				Genode::network_profile(SAMPLING_SUSPEND, &tid_spec, network_buf);
//			else
				L4_ProfPolicy(SAMPLING_SUSPEND, &tid_spec);
		}
	}
}

void resume(ProfVector<L4_ThreadId_t> *vec)
{
//	L4_Word_t tid_search = thread_ptr_index((uintptr_t)fn);
	buffer_info *tbuffer;// = thread_buffer_info(tid_search);
	
	int i;
	for(i = 0;i < vec->length;i++)
	{
		tbuffer = binfo_raw_thread(vec->buffer[i].raw);
	
		if(tbuffer->interrupt_no == 0)
		{
			//it is not a probe, so use the default scheduler
			L4_ThreadId_t thr_resume;
			thr_resume.raw = vec->buffer[i].raw;
			L4_Start(thr_resume);
		}
		else
		{
			interrupt_spec tid_spec;
			tid_spec.tid = vec->buffer[i].raw;
			int tu_assigned = 0;
			
			switch(tbuffer->interrupt_no)
			{
				case MSEC_TIMER:
				{
					if(thread_tunit[TID_MSEC].raw == 0)
					{
						assign_tid_tunit(vec->buffer[i], TID_MSEC);
						tu_assigned = 1;
					}
					else
						PDBG("MSEC PROBE IS CURRENTLY BEING USED\n");
					break;
				}
				case SEC_TIMER:
				{
					if(thread_tunit[TID_SEC].raw == 0)
					{
						assign_tid_tunit(vec->buffer[i], TID_SEC);
						tu_assigned = 1;
					}
					else
						PDBG("SEC PROBE IS CURRENTLY BEING USED\n");
					break;
				}
				case MIN_TIMER:
				{
					if(thread_tunit[TID_MIN].raw == 0)
					{
						assign_tid_tunit(vec->buffer[i], TID_MIN);
						tu_assigned = 1;
					}
					else
						PDBG("MIN PROBE IS CURRENTLY BEING USED\n");
					break;
				}
				case HR_TIMER:
				{
					if(thread_tunit[TID_HR].raw == 0)
					{
						assign_tid_tunit(vec->buffer[i], TID_HR);
						tu_assigned = 1;
					}
					else
						PDBG("HR PROBE IS CURRENTLY BEING USED\n");
					break;
				}
				case PMC1:
				{
					if(thread_tunit[TID_PMC1].raw == 0)
					{
						assign_tid_tunit(vec->buffer[i], TID_PMC1);
						tu_assigned = 1;
					}
					else
						PDBG("PMC1 PROBE IS CURRENTLY BEING USED\n");
					break;
				}				
				case PMC2:
				{
					if(thread_tunit[TID_PMC2].raw == 0)
					{
						assign_tid_tunit(vec->buffer[i], TID_PMC2);
						tu_assigned = 1;
					}
					else
						PDBG("PMC2 PROBE IS CURRENTLY BEING USED\n");
					break;
				}				
				case HTTP_SERVER:
				{
					if(thread_tunit[TID_HTTP].raw == 0)
					{
						assign_tid_tunit(vec->buffer[i], TID_HTTP);
						tu_assigned = 1;
					}
					else
						PDBG("HTTP PROBE IS CURRENTLY BEING USED\n");
					break;
				}
			}
			
			if(tu_assigned)
			{
//				if(tbuffer->interrupt_no == HTTP_SERVER)
//					Genode::network_profile(SAMPLING_RESUME, &tid_spec, network_buf);
//				else //assign the tid unit to the resumed probe
					L4_ProfPolicy(SAMPLING_RESUME, &tid_spec);
			}
		}
	}
}

void kill(ProfVector<L4_ThreadId_t> *vec)
{
	//L4_Word_t tid_search = thread_ptr_index((uintptr_t)fn);

	int i;
	buffer_info *tbuffer;
	
	for(i = 0;i < vec->length;i++)
	{
		tbuffer = binfo_raw_thread(vec->buffer[i].raw);
	
		if(tbuffer->interrupt_no == 0)
		{
			//it is not a probe, so use the default scheduler
			L4_ThreadId_t thr_kill;
			thr_kill.raw = vec->buffer[i].raw;

			L4_ThreadControl(thr_kill, L4_nilspace, L4_nilthread, L4_nilthread, L4_nilthread, 0, NULL);
		}
		else
		{
			interrupt_spec tid_spec;
			tid_spec.tid = vec->buffer[i].raw;
			
			switch(tbuffer->interrupt_no)
			{
				case MSEC_TIMER:
				{
					if(thread_tunit[TID_MSEC].raw == vec->buffer[i].raw)
						thread_tunit[TID_MSEC].raw = 0;
						
					break;
				}
				case SEC_TIMER:
				{
					if(thread_tunit[TID_MSEC].raw == vec->buffer[i].raw)
						thread_tunit[TID_SEC].raw = 0;
					break;
				}
				case MIN_TIMER:
				{
					if(thread_tunit[TID_MSEC].raw == vec->buffer[i].raw)
						thread_tunit[TID_MIN].raw = 0;
					break;
				}
				case HR_TIMER:
				{
					if(thread_tunit[TID_MSEC].raw == vec->buffer[i].raw)
						thread_tunit[TID_HR].raw = 0;
					break;
				}
				case PMC1:
				{
					if(thread_tunit[TID_MSEC].raw == vec->buffer[i].raw)
						thread_tunit[TID_PMC1].raw = 0;
					break;
				}
				case PMC2:
				{
					if(thread_tunit[TID_MSEC].raw == vec->buffer[i].raw)
						thread_tunit[TID_PMC2].raw = 0;
					break;
				}
				case HTTP_SERVER:
				{
					if(thread_tunit[TID_MSEC].raw == vec->buffer[i].raw)
						thread_tunit[TID_HTTP].raw = 0;
					break;
				}
			}
			
//			if(tbuffer->interrupt_no == HTTP_SERVER)
//			{	
//				Genode::network_profile(SAMPLING_KILL, &tid_spec, network_buf);
//			}
//			else
				L4_ProfPolicy(SAMPLING_KILL, &tid_spec);

		}
	}
}


void register_thread(L4_ThreadId_t tid_to_register)//, void (*fn)(void*))
{
	L4_MsgTag_t tag;
	L4_Msg_t msg;
	L4_MsgClear(&msg);

	L4_MsgAppendWord(&msg, tid_to_register.raw);
	//L4_MsgAppendWord(&msg, (uintptr_t)fn);
	L4_MsgAppendWord(&msg, 0);
	L4_MsgLoad(&msg);
	
	tag = L4_Send(register_id);
}



//void new_input_link(void (*fn1)(void*), void (*fn2)(void*))
void new_input_link(L4_ThreadId_t my_thread, buffer_info *input_data)
{
	L4_MsgTag_t tag;
	L4_Msg_t msg;
	L4_MsgClear(&msg);
	
	//L4_MsgAppendWord(&msg, thread_ptr_index((uintptr_t)fn1));
	//L4_MsgAppendWord(&msg, thread_ptr_index((uintptr_t)fn2));
	L4_MsgAppendWord(&msg, thread_index(my_thread));
	L4_MsgAppendWord(&msg, buffer_info_index(input_data));
	L4_MsgLoad(&msg);

	tag = L4_Send(matrix_id);
}

int thread_index_raw(L4_Word_t input)
{
	for(int i = 0;i < matrix_index;i++)
	{
		if(thread_list[i] == input)
			return i;
	}
	
	return -1;
}

/* 	
	As part of the thread registration, the first step is to
	enter a reference to that thread in the thread entry array
	
	Parameters:
		- sampler: the ID of the thread to register
		- buffer_addr: the address of the thread's sampling buffer
*/
L4_Word_t enter_thread_entry(L4_Word_t tid_register, buffer_info *buffer)//uintptr_t fn_ptr, buffer_info *buffer)
{	
	int tindex = thread_index_raw(tid_register);
	//int tindex = thread_ptr_index(fn_ptr);
	if(tindex != -1)
	{
		thread_buffers[tindex] = buffer;
		return tindex;
	}
	else if(matrix_index <= THREAD_QUANTITY)
	{
		thread_buffers[matrix_index] = buffer;
		thread_list[matrix_index] = tid_register;
		//thread_ptrs[matrix_index] = fn_ptr;
		matrix_index++;
		return matrix_index - 1;
	}
	
	return -1;
}

/*	May have to change it so that it is implied that the thread that called it wants to
	create the link to another thread */
void enter_matrix_entry(int source_thread_index, int destination_thread_index)
{

	/* 	Ensure that the thread we want to connect to actually
		has its own buffer info structure */
	if(thread_buffer_info(destination_thread_index) == NULL)
	{
		PDBG("destination_thread_index: Invalid thread index specified\n");
		return;
	}
	
	/*	Ensure the input thread is actually valid.  This is checked by
		seeing whether the index is less than the total number of sampling
		threads */
	if(source_thread_index >= matrix_index)
	{
		PDBG("source_thread_index: Invalid thread index specified\n");
		return;
	}

	//Here, we want to store a reference to the index of the buffer.
	//ATM, need to find a more reliable way of doing this though.
	//Also, the format of the matrix is: output, input ie. row, column.
	//That is, the rows correspond to the threads looking at a thread's output
	//while the columns indicate the inputs that the thread is currently using.
	thread_matrix[destination_thread_index][source_thread_index] = 1;
	buffer_info *thread_buffer = thread_buffer_info(destination_thread_index);

	thread_buffer->threads[source_thread_index] = thread_buffer->buffer;
	thread_buffer->connected_threads++;
}

void min_time_unit(buffer_info *info, int *curr_unit, L4_Word_t tindex)
{
	buffer_info *next_buffer;
	int iterate = 0;
	int found = 0;
	int index = tindex;
	
	if(index == -1)
	{
		L4_Stop_Thread(L4_Myself());
	}
	
	while(iterate < matrix_index)
	{
		if(thread_matrix[iterate][index] == 1)
		{
			found = 1;
			next_buffer = thread_buffer_info(iterate);
			assert(next_buffer != NULL);
			if(next_buffer->storage_mode == CONSOLIDATE)
			{
				if(next_buffer->window_unit < info->window_unit)
				{
					*curr_unit = next_buffer->window_unit;
				}
			}
			
			//if the thread being connected to is not consolidating data at all,
			//the traversal simply continues.  Eventually, there will be "nodes"
			//that will have a time unit.  These correspond to the collection
			//specifications.
			min_time_unit(next_buffer, curr_unit, iterate);
		}
		
		iterate++;
	}
	
	if(found == 0)
	{	next_buffer = thread_buffer_info(index);
		if(next_buffer->interrupt_no < *curr_unit)
		{
			*curr_unit = info->interrupt_no;
		}
	}
}

int window_factor(buffer_info *info)
{
	int tunit_analysis;
	if(info->storage_mode == CONSOLIDATE)
		tunit_analysis = info->window_unit;
	else
		tunit_analysis = info->interrupt_no;

	int unit = tunit_analysis;
	min_time_unit(info, &unit, thread_index(L4_Myself()));
	
	if((tunit_analysis == UNIT_MSEC && unit == UNIT_MSEC) ||
		(tunit_analysis == UNIT_SEC && unit == UNIT_SEC) ||
		(tunit_analysis == UNIT_MIN && unit == UNIT_MIN) ||
		(tunit_analysis == UNIT_HR && unit == UNIT_HR))
	{
		return 1;
	}
	else if(tunit_analysis == UNIT_MSEC)
	{		
		if(unit == UNIT_SEC || unit == UNIT_MIN || unit == UNIT_HR)
		{
			PDBG("UNIT_TICK: Cannot provide consolidation based on units\n");
			return -1;
		}
	}
	else if(tunit_analysis == UNIT_SEC)
	{
		if(unit == UNIT_MSEC)
			return 513;
		else if(unit == UNIT_MIN || unit == UNIT_HR)
		{
			PDBG("UNIT_SEC: Cannot provide consolidation based on units\n");
			return -1;
		}
	}
	else if(tunit_analysis == UNIT_MIN)
	{
		if(unit == UNIT_SEC)
			return 60;
		else if(unit == UNIT_MSEC)
			return 60 * 513;
		else if(unit == UNIT_HR)
		{
			PDBG("UNIT_MIN: Cannot consolidate based on units\n");	
			return -1;
		}
	}
	else if(tunit_analysis == UNIT_HR)
	{
		if(unit == UNIT_MSEC)
			return 513 * 60 * 60;
		else if(unit == UNIT_SEC)
			return 60 * 60;
		else if(unit == UNIT_MIN)
			return 60;
	}
	
	return -1;
}

L4_Word_t window_constant(buffer_info *info)
{//this should only apply to a collection spec
	/*
		If the time units for the sample collection and consolidation are the same,
		the amount of memory to allocate is always going to be:
		
		window length * sample size * number of windows
	*/
	if((info->interrupt_no == MSEC_TIMER && info->window_unit == UNIT_MSEC)
		|| (info->interrupt_no == SEC_TIMER && info->window_unit == UNIT_SEC)
		|| (info->interrupt_no == MIN_TIMER && info->window_unit == UNIT_MIN)
		|| (info->interrupt_no == HR_TIMER && info->window_unit == UNIT_HR))
	{
		return 1;
	}
	else if(info->interrupt_no == MSEC_TIMER)
	{
		if(info->window_unit == UNIT_SEC)
			return 513;
		else if(info->window_unit == UNIT_MIN)
			return 30780;
		else if(info->window_unit == UNIT_HR)
			return 1846800;
	}
	else if(info->interrupt_no == SEC_TIMER)
	{
		if(info->window_unit == UNIT_MSEC)
			return -1; 
		else if(info->window_unit == UNIT_MIN)
			return 60;
		else if(info->window_unit == UNIT_HR)
			return 3600;
	}
	else if(info->interrupt_no == MIN_TIMER)
	{
		if(info->window_unit == UNIT_MSEC || info->window_unit == UNIT_SEC)
			return -1;
		else if(info->window_unit == UNIT_HR)
			return 60;
	}
	else if(info->interrupt_no == HR_TIMER)
	{
		if(info->window_unit == UNIT_MSEC || info->window_unit == UNIT_SEC || info->window_unit == UNIT_HR)
			return -1;
	}
	
	return -1;
}

//size: refers to the quantity of types to be used for the buffer
void initialize_buffer(interrupt_spec *int_info, buffer_info *info, char *type_name, L4_ThreadId_t tid)//void (*fn)(void*))
{
	if(int_info != NULL)
	{
		info->interrupt_no = int_info->interrupt_no;
		info->info_collect = int_info->info_collect;
		info->duration = int_info->duration;
		info->rate = int_info->rate;
		info->time_window =int_info->time_window;
		info->window_unit = int_info->window_unit;
		info->no_windows = int_info->no_windows;
		info->storage_mode = int_info->storage_mode;
	
		switch(info->type_identifier)
		{
			case -4: //boolean
				info->actual_chunk_size = 1;
				break;
			case -3: //string
				info->actual_chunk_size = 50;
				break;
			case -2: //double
				info->actual_chunk_size = sizeof(double);
				break;
			case -1:
				info->actual_chunk_size = sizeof(int);
				break;
			case 0:
				info->actual_chunk_size = sizeof(int) * 2;
				break;
			case 1:
				info->actual_chunk_size = sizeof(int) * 3;
				break;
			case 2:
				info->actual_chunk_size = sizeof(int) * 4;
				break;
			default:
			{
				tinfo *info_about_type = get_stats(info->type_identifier);
				info->actual_chunk_size = memory_single_chunk((char*)info_about_type->tup_type_seq) + sizeof(int);//additional id field
			}
		}
	}
	
	int index_thread = thread_index_raw(tid.raw);//fn(fn);
	buffer_info *iter;
	int found = 0;
	
	int i;
	for(i = 0;i < matrix_index;i++)
	{//Allocation of memory for all inputs to this thread.  This checks to see whether
	 //there are inputs to the thread so that the memory can be allocated correctly.
		if(thread_matrix[index_thread][i] == 1)
		{
			found = 1;
			iter = thread_buffer_info(i);
			
			//if i'm consolidating, look at the windowing information.  Check compatibility
			//and allocate the appropriate amount of memory.  Before this, consolidation modes
			//must be compatible as well ie. if the thread being connected to is not consolidating,
			//logic would suggest that this thread shouldn't be in consolidation mode as there is no
			//notion of windowing unconsolidated data.
			if(iter->storage_mode == CONSOLIDATE)
			{
				if(info->window_unit == UNIT_SEC && info->interrupt_no == MSEC_TIMER)
				{
					if(info->no_windows < 7)
					{
						info->no_windows = 7;
						int_info->no_windows = 7;
					}
				}
			}
			else if(info->storage_mode == CONSOLIDATE)
			{
				//if so, we cannot consolidate data from a thread that is not
				//consolidating data, so change the mode appropriately
				info->storage_mode == NO_CONSOLIDATE;
			}

			info->remaining_size = info->actual_chunk_size;
			
			if(int_info != NULL)
			{
				//Logic would suggest that if the thread is consolidating, the thread being
				//connected to is also consolidating.  This will be checked upon making a thread
				//link
				if(info->storage_mode == CONSOLIDATE)
				{	
					//Get the required memory for the consolidation buffer.  Calculations based on
					//"worst case" scenario
					if(info->window_unit == UNIT_SEC && info->interrupt_no == MSEC_TIMER)
					{
						if(info->no_windows < 7)
						{
							info->no_windows = 7;
							int_info->no_windows = 7;
						}
					}
					
					int win_factor = window_factor(info);
					info->time_window *= win_factor;
					
					if(info->time_window < iter->time_window)
					{
						info->time_window = iter->time_window;
					}

					while(info->time_window % iter->time_window != 0)
					{
						info->time_window++;
					}

					L4_Word_t new_size = iter->core_size * 
										 (info->time_window / iter->time_window) *
										 info->no_windows;

					info->core_size = iter->core_size * 
										 (info->time_window / iter->time_window);
					info->buffer = realloc((char*)info->buffer, info->size, info->size + new_size + sizeof(char));
					info->size += new_size + sizeof(char);
					info->write_index = info->buffer;
				}
				else if(iter->storage_mode == CONSOLIDATE)
				{//If the thread we are connecting to is consolidating, we should at the very least allocate
				 //data for the sampling window.
				 
					info->buffer = realloc((char*)info->buffer, info->size, iter->size);
					info->size += iter->size;
					info->write_index = info->buffer;
				}
				else
				{
					/* allocate a fixed amount of memory -> 130KB thereabouts => half of one buffer */
					info->buffer = realloc((char*)info->buffer, info->size, info->size + 131048);
					info->size += 131048;
					info->write_index = info->buffer;
				}
			}
			else
			{
				PDBG("Interrupt information not supplied\n");
				L4_Stop_Thread(L4_Myself());
			}
		}
	}
	
	if(found == 0)
	{//This must be a collection specification, so allocate memory based on the interrupt
	 //information supplied.
		if(info->storage_mode == CONSOLIDATE)
		{
			//PDBG("About to create consolidated memory. SO I KNOW IT'S INITTING\n");
			L4_Word_t consol_size = determine_size(info);
			//PDBG("consolidation size: %ld\n", consol_size);
			
			/*
				memory for storing count data is already used in buffer calculations
			*/
			//(char*)malloc(consol_size + int_info->no_windows + sizeof(char));
			info->buffer = new(env()->heap())char[consol_size + int_info->no_windows + sizeof(char)];
			info->write_index = info->buffer;
			info->size = consol_size + int_info->no_windows + sizeof(char);
						 
			/*
				core_samples -> time window, including additional field
				
				size of a single window:
				consol_size/number of windows + (number of possible records in a window * size of field)
			*/
			
			//what i want to do here is only collect the required information in the first buffer. The
			//intermediate buffer will then have the required memory capacity (including the additional
			//fields for type id and count
			//(char*)malloc(consol_size / int_info->no_windows - (window_constant(info) * 8) + sizeof(char));
			info->core_samples = new(env()->heap())char[consol_size / int_info->no_windows - (window_constant(info) * 8) + sizeof(char)];
			info->core_size = consol_size / int_info->no_windows - (window_constant(info) * 8) + sizeof(char);
			info->core_pointer = info->core_samples;
		}
		else
		{
			//(char*)malloc(131055)
			info->buffer = new(env()->heap())char[131055];
			info->size = 131055;
			info->write_index = info->buffer;
		}
	}
	
	L4_MsgTag_t tag;
	L4_Msg_t msg;
	L4_MsgClear(&msg);
	L4_MsgAppendWord(&msg, tid.raw);
	//L4_MsgAppendWord(&msg, (uintptr_t)fn);
	L4_MsgAppendWord(&msg, (L4_Word_t)info);
	
	L4_MsgLoad(&msg);
	
	//A message is sent to the matrix thread for registration.

	tag = L4_Send(register_id);
	
}

void initialize_binfo(buffer_info *info)
{
	info->interrupt_no = 0;
	info->info_collect = 0;
	info->thread_index = 0;
	info->duration = 0;
	info->rate = 0;
	info->time_window = 0;
	info->window_unit = 0;
	info->no_windows = 0;

	info->storage_mode = NO_CONSOLIDATE;
	info->core_size = 0;

	int i;
	for(i = 0;i < THREAD_QUANTITY;i++)
	{
		info->threads[i] = NULL;
		info->moved[i] = 0;
		info->resetted[i] = 0;
		info->t_offsets[i] = 0;
	}
	
	info->size = 0;
	info->reset = 0;
	
	info->connected_threads = 0;
	info->no_scanned = 0;
	
	info->type_identifier = -5;
	
	info->dropped_samples = 0;
	info->actual_chunk_size = 0;
	info->remaining_size = info->chunk_size;
}

void collect_samples(int sampling_params, interrupt_spec *int_info, buffer_info *info, L4_ThreadId_t tid)//void (*fn)(void*))
{
	//See if the buffer has memory allocated.  If not, allocate the memory
	if(info->buffer == NULL)
	{
		L4_MsgTag_t tag;
		L4_Msg_t msg;	
		L4_MsgClear(&msg);
			
		initialize_binfo(info);
		
		if(int_info->info_collect & INFO_TID &&
			int_info->info_collect & INFO_INST)
		{
			if(int_info->storage_mode == CONSOLIDATE)
			{
				info->type_identifier = 2;
			}
			else
			{
				info->type_identifier = 1;
			}
		}
		else if(int_info->info_collect & INFO_TID ||
				int_info->info_collect & INFO_INST)
		{
			if(int_info->storage_mode == CONSOLIDATE)
			{
				info->type_identifier = 1;
			}
			else
			{
				info->type_identifier = 0;
			}
		}
		else
		{
			PDBG("INITIALIZE BUFFER: Collection parameters not specified properly\n");
			return;
		}
		
		initialize_buffer(int_info, info, NULL, tid);
		
		tag = L4_Receive(register_id);
		L4_MsgStore(tag, &msg);
		
		int given_index = (int)L4_MsgWord(&msg, 0);
		assert(given_index == thread_index_raw(tid.raw));
	}
	
	switch(int_info->interrupt_no)
	{
		case MSEC_TIMER:
			assign_tid_tunit(tid, TID_MSEC);
			break;
		case SEC_TIMER:
			assign_tid_tunit(tid, TID_SEC);
			break;
		case MIN_TIMER:
			assign_tid_tunit(tid, TID_MIN);
			break;
		case HR_TIMER:
			assign_tid_tunit(tid, TID_HR);
			break;
		case PMC1:
			assign_tid_tunit(tid, TID_PMC1);
			break;
		case PMC2:
			assign_tid_tunit(tid, TID_PMC2);
			break;
		case HTTP_SERVER:
			assign_tid_tunit(tid, TID_HTTP);
			break;
	}

//	if(int_info->interrupt_no == HTTP_SERVER)
//	{
//		Genode::network_profile(sampling_params, int_info);
//	}
//	else
//	{
		int_info->duration = int_info->duration * duration_factor(int_info->rate_unit, int_info->duration_unit);
		//Perform the system call to collect samples
		L4_ProfPolicy(sampling_params, int_info);
//	}
}

int duration_factor(L4_Word_t rate_unit, L4_Word_t duration_unit)
{
	if(rate_unit == UNIT_MSEC)
	{
		if(duration_unit == UNIT_MSEC)
			return 1;
		else if(duration_unit == UNIT_SEC)
			return 1000000;
		else if(duration_unit == UNIT_MIN)
			return 1000000 * 60;
		else if(duration_unit == UNIT_HR)
			return 3600000000;
	}
	else if(rate_unit == UNIT_SEC)
	{
		if(duration_unit == UNIT_MSEC)
			return 0;
		else if(duration_unit == UNIT_SEC)
			return 1000000;
		else if(duration_unit == UNIT_MIN)
			return 60000000;
		else if(duration_unit == UNIT_HR)
			return 3600000000;
	}
	else if(rate_unit == UNIT_MIN)
	{
		if(duration_unit == UNIT_MSEC || duration_unit == UNIT_SEC)
			return 0;
		else if(duration_unit == UNIT_MIN)
			return 60000000;
		else if(duration_unit == UNIT_HR)
			return 3600000000;
	}
	else if(rate_unit == UNIT_HR)
	{
		if(duration_unit == UNIT_MSEC || duration_unit == UNIT_SEC || duration_unit == UNIT_MIN)
			return 0;
		else
			return 3600000000;
	}
	return 0;
}

char *get_record(char *buff_ptr, int size)
{
	char *new_record = new(env()->heap())char[size];
	Genode::memcpy(new_record, buff_ptr, size);
	return new_record;
}

void finish_stream(L4_ThreadId_t tid)
{
	buffer_info *info = binfo_raw_thread(tid.raw);
	//buffer_info *info = binfo_fn((uintptr_t)fn);
	L4_Word_t min_index = minimum_index_bounds(info);
	int first_time = 0;
	
	if(info->reset == 0)
	{
		if(info->write_index + 4 >= info->buffer + info->size)
		{
			if(min_index != NO_THREAD_LINKS && min_index != NO_BUFFER_SCANS &&
				info->buffer + 4 < info->buffer + min_index)
			{
				first_time = 1;
			}

			set_reset_state(info);
			info->reset = 1;
		}
		else
		{
			printf("END OF RECORD WRITTEN\n");
			char *end = (char*)"END";
			memcpy((void*)info->write_index, end, strlen(end) + 1);
			info->write_index += 4;
		}
	}
	
	if(info->reset == 1)
	{
		if(first_time == 1)
		{
			first_time = 0;
			*(info->write_index) = BUFFER_RESET;
			info->write_index = info->buffer;
		}
		else if(info->write_index + 1 < info->buffer + min_index)
		{
			//*(info->write_index) = END_OF_RECORDS;
			memcpy((void*)info->write_index, "END", strlen("END") + 1);
			info->write_index += 4;
		}
		else
		{
			printf("about to drop a sample\n");
			info->dropped_samples++;
		}
	}
}


/**************************** CREATE THREADS FOR THE FOLLOWING *******************************************/


#if 0

#endif
